/*
 * Copyright (C) 2017-2021 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "backoff-flooding-ring-routing-agent.h"
#include "../../scheduler.h"


//==============================================================================
//
//          BackoffFloodingDynamicRingRoutingAgent    (class)
//
//==============================================================================
set<int> BackoffFloodingDynamicRingRoutingAgent::reachability = set<int>();
mt19937_64 *BackoffFloodingDynamicRingRoutingAgent::forwardingRNG;
int BackoffFloodingDynamicRingRoutingAgent::redundancy;
float BackoffFloodingDynamicRingRoutingAgent::dataBackoffMultiplier;


BackoffFloodingDynamicRingRoutingAgent::BackoffFloodingDynamicRingRoutingAgent (Node *_hostNode) : RoutingAgent(_hostNode) {
  needToSendControl = true;
  rangeBig = ScenarioParameters::getRangeBig();
  //rangeSmall = ScenarioParameters::getRangeSmall();
  initialisationStarted = false;
}

BackoffFloodingDynamicRingRoutingAgent::~BackoffFloodingDynamicRingRoutingAgent () {
  aliveAgents--;

  if (aliveAgents == 0) {
    //        TODO : Info globale backoff Flooding
    cout << " Backoff flooding with ring:: number of nodes reached : " << reachability.size() << endl;
    cout << " Backoff flooding with ring:: forwarded packets       : " << forwardedDataPackets << endl;

    //     string filename = ScenarioParameters::getScenarioDirectory()+"/reachability_cost.data";
    //     ofstream file(filename,std::ofstream::app);
    //     file << forwardedDataPackets << " " ;
    //     file.close();
  }
}

void BackoffFloodingDynamicRingRoutingAgent::initializeAgent() {
  forwardingRNG =  new mt19937_64(ScenarioParameters::getBackoffRNGSeed());
  redundancy = ScenarioParameters::getBackoffRedundancy();
  dataBackoffMultiplier = ScenarioParameters::getBackoffMultiplier();
  readRangeAttributes ();
}

void BackoffFloodingDynamicRingRoutingAgent::receivePacketFromApplication(PacketPtr _packet) {

  if ( _packet->type == PacketType::DATA ) {
    float optimal_NbRingNeighborsPerHop = 110.0;
    if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

      //rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));
	
      //journal if 3D??
      rangeSmall = pow(((1 - optimal_NbRingNeighborsPerHop/hostNode->getEstimatedNeighbours()) * pow(ScenarioParameters::getRangeBig(),3)),1.0/3);

    else rangeSmall = 0;
     
    cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

    if (needToSendControl) { //1 and 2 before the very first data transmission
      PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
      controlBig->setCommunicationRange(rangeBig);
      controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlBig);
      PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
      controlSmall->setCommunicationRange(rangeSmall);
      controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlSmall);	
      needToSendControl = false;
    }

    _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
  }
  _packet->disableBackoff = true;
  cout << hostNode->getId() << " onring of " << amIOnRingCount() << " transmitters" << endl; //journal
  hostNode->enqueueOutgoingPacket(_packet);

  if ( _packet->type == PacketType::DATA ) {
    datapacket_t datapacket;
    datapacket.first =  _packet->srcId;
    datapacket.second = _packet->srcSequenceNumber;    
    auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
    if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
      auto timerEvent = new BackoffFloodingDynamicRingAckEvent(Scheduler::now(), this, _packet); 
      Scheduler::getScheduler().schedule(timerEvent);
    }
  }
}

void BackoffFloodingDynamicRingRoutingAgent::receivePacketFromNetwork(PacketPtr _packet){

  LogSystem::EventsLogOutput.log( LogSystem::routingRCV, Scheduler::now(),hostNode->getId(),(int) _packet->type,_packet->flowId,_packet->flowSequenceNumber, _packet->modifiedBitsPositions.size());
  //   bool insertedPackets = false;
  //   bool erasedPackets = false;

  // source must not resend the packet
  //if ( _packet->srcId == hostNode->getId())
  //return;

  if ( _packet->modifiedBitsPositions.size() > 0 ) {
    return;
  }

  if (_packet->type == PacketType::DATA){
    auto srcIt = alreadySeenPackets.find(_packet->srcId);
    if (srcIt == alreadySeenPackets.end() || srcIt->second < _packet->srcSequenceNumber){
      if (srcIt != alreadySeenPackets.end()) {  // already seen
	if (srcIt->second < _packet->srcSequenceNumber) {
	  // the source sequence number of this packet is higher than the highest already seen from this source
	  srcIt->second = _packet->srcSequenceNumber;
	}
      } 
      else {  // seen for the first time
	alreadySeenPackets.insert(pair<int,int>(_packet->srcId, _packet->srcSequenceNumber));
      }
      if(amIOnRing(_packet)){
	waitingDataPackets(_packet);
      }
    }//if not already seen

    // If I already saw this packet REDUNDANCY or more times, I prevent any retransmission already scheduled
    //
    // Also further copies of this packet will not be retransmitted either, as
    // alreadySeen list should prevent it.
    //
    for (auto it = waitingPacket.begin(); it!=waitingPacket.end();it++){
      if ( (it->second.p)->flowId == _packet->flowId && (it->second.p)->flowSequenceNumber == _packet->flowSequenceNumber) {
	if (it->second.counter >= 2){ //redundancy
	  it = waitingPacket.erase(it);
	  // erasedPackets=true;
	  // LogSystem::EventsLogOutput.log( LogSystem::memoryTrace, Scheduler::now(),hostNode->getId(),"-",_packet->flowId,_packet->flowSequenceNumber);
	} else {
	  it->second.counter++;
	} 
	break;
      }
    }
  }//if data
  
  //CONTROL_BIG
  else if (_packet->type == PacketType::CONTROL_1) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// transmiter already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoBig = _packet->ctrl1SequenceNumber;
    } 
    else {
      // transmitter seen for the first time
      contrSeqNo_t control;
      control.controlSeqNoBig = _packet->ctrl1SequenceNumber;
      control.controlSeqNoSmall = -1;
      amIonRingMap.insert(pair<int,contrSeqNo_t>(_packet->transmitterId, control));
    }	
  }

  //CONTROL_SMALL
  else if (_packet->type == PacketType::CONTROL_2) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
    } 
    else {
      // seen for the first time
      contrSeqNo_t control;
      control.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
      control.controlSeqNoBig = -1;
      amIonRingMap.insert(pair<int,contrSeqNo_t>(_packet->transmitterId, control));	
    }	
  }

  else if ( _packet->type == PacketType::D1_DENSITY_PROBE) {  //PROBE is local
    hostNode->dispatchPacketToApplication(_packet);
    return;
  }

  else if (  _packet->type == PacketType::D1_DENSITY_INIT && !initialisationStarted ) { //INIT is global
    initialisationStarted = true;
    hostNode->dispatchPacketToApplication(_packet);
    PacketPtr newPacket(_packet->clone());
    hostNode->enqueueOutgoingPacket(newPacket);
    return;
  }

  if (_packet->type == PacketType::DATA && _packet->dstId == hostNode->getId() ) {
    hostNode->dispatchPacketToApplication(_packet);
    reachability.insert(hostNode->getId());
    LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber);
    return;
  }

  if (  _packet->type == PacketType::DATA || _packet->type == PacketType::CONTROL_1 || _packet->type == PacketType::CONTROL_2){
    if (_packet->dstId == -1) {
      hostNode->dispatchPacketToApplication(_packet);
      LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber);
      //       reachability.insert(hostNode->getId());
      //       return;
    }
  }
}

//amIOnRing
bool BackoffFloodingDynamicRingRoutingAgent::amIOnRing( PacketPtr _packet ) {
  auto trIt = amIonRingMap.find(_packet->transmitterId);
  if (trIt == amIonRingMap.end()) 
    return false;

  else {
    //cout << "Farah big " << trIt->second.controlSeqNoBig << " small " << trIt->second.controlSeqNoSmall << endl; //journal
    if ((trIt->second.controlSeqNoBig > trIt->second.controlSeqNoSmall) || (trIt->second.controlSeqNoBig != -1 && trIt->second.controlSeqNoSmall == -1)){
      //cout << hostNode->getId() << " onringof " << _packet->transmitterId << endl;
      //cout << hostNode->getId() << " onring of " << amIonRingMap.size() << " transmitters" << endl; //journal
      return true;}
    else
      return false;
  }
}

//amIOnRingCount
int BackoffFloodingDynamicRingRoutingAgent::amIOnRingCount() {
	int i =0;
	for(map<int,contrSeqNo_t> ::const_iterator itt = amIonRingMap.begin(); itt != amIonRingMap.end(); ++itt)
		    {
		  if ((itt->second.controlSeqNoBig > itt->second.controlSeqNoSmall) || (itt->second.controlSeqNoBig != -1 && itt->second.controlSeqNoSmall == -1))
		  i++;
		    }
	return i;
}

//waitingDataPackets
void BackoffFloodingDynamicRingRoutingAgent::waitingDataPackets( PacketPtr _packet ) {
  simulationTime_t backoff = getBackoffDynamicRing(hostNode, _packet,
						   _packet->type == PacketType::DATA ? dataBackoffMultiplier : 1,
						   forwardingRNG);
  PacketPtr packetClone(_packet->clone());
  struct backoffRoutingCounter info;
  info.p = packetClone;
  info.counter = 1;
  waitingPacket.insert(pair<int,backoffRoutingCounter>(packetClone->packetId,info));
  //             insertedPackets = true;
  Scheduler::getScheduler().schedule(new backoffDynamicRingSendingEvent(Scheduler::now() + backoff, hostNode,packetClone->packetId));
  //             LogSystem::EventsLogOutput.log( LogSystem::memoryTrace, Scheduler::now(),hostNode->getId(),"+",_packet->flowId,_packet->flowSequenceNumber);
}


void BackoffFloodingDynamicRingRoutingAgent::processBackoffSending(int packetId){
  auto it = waitingPacket.find(packetId);
  if ( it != waitingPacket.end() && it->second.p->type == PacketType::DATA ) {
    //
    // if the packet is still in the waiting list
    // i.e. I have not see more than REDUNDANCY copies
    // then I send it now
    //
    //cout << " ++ forward seq " << it->second.p->flowSequenceNumber << " by node " << hostNode->getId()<< " at " << Scheduler::getScheduler().now() << endl;

  if (hostNode->getId() == 0) return;

    float optimal_NbRingNeighborsPerHop = 110.0;
    if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

      //rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));
	
      //journal if 3D??
      rangeSmall = pow(((1 - optimal_NbRingNeighborsPerHop/hostNode->getEstimatedNeighbours()) * pow(ScenarioParameters::getRangeBig(),3)),1.0/3);

    else rangeSmall = 0;
     
    cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

    if (needToSendControl) { //before the very first data transmission
      PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
      controlBig->setCommunicationRange(rangeBig);
      controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlBig);
      PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
      controlSmall->setCommunicationRange(rangeSmall);
      controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlSmall);	
      needToSendControl = false;													     
    }

		cout << hostNode->getId() << " onring of " << amIOnRingCount() << " transmitters" << endl; //journal
  /*if (hostNode->getId() == 5851){
		  for(map<int,contrSeqNo_t> ::const_iterator itt = amIonRingMap.begin();
		itt != amIonRingMap.end(); ++itt)
		    {
		std::cout << " Eugen "  << itt->first << " " << itt->second.controlSeqNoBig << " " << itt->second.controlSeqNoSmall << "\n";
		    }
      }*/
    hostNode->enqueueOutgoingPacket(it->second.p);
    LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) it->second.p->type,it->second.p->flowId,it->second.p->flowSequenceNumber);
    //         LogSystem::EventsLogOutput.log( LogSystem::memoryTrace, Scheduler::now(),hostNode->getId(),"-",it->second.p->flowId,it->second.p->flowSequenceNumber);

    forwardedDataPackets++;
    waitingPacket.erase(it);
    datapacket_t datapacket;
    datapacket.first =  it->second.p->srcId;
    datapacket.second = it->second.p->srcSequenceNumber;    
    auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
    if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
      auto timerEvent = new BackoffFloodingDynamicRingAckEvent(Scheduler::now(), this, it->second.p); 
      Scheduler::getScheduler().schedule(timerEvent);
    }       
    //       string filename = ScenarioParameters::getScenarioDirectory()+"/dataSent.data";
    //       ofstream dataSent(filename,std::ofstream::app);
    //       dataSent<< ScenarioParameters::getDefaultBeta() << " " <<hostNode->getId() << " "<< Scheduler::now() << endl;
    //       dataSent.close();
    //
    //
    //       string filename2 = ScenarioParameters::getScenarioDirectory()+"/memoryTrace.data";
    //       ofstream memoryTrace(filename2,std::ofstream::app);
    //       memoryTrace << ScenarioParameters::getDefaultBeta() <<" " << hostNode->getId() <<  " - "<< Scheduler::now() << endl;
    //       memoryTrace.close();
  }
}

//processAckTimerEvent
void BackoffFloodingDynamicRingRoutingAgent::processAckTimerEvent(PacketPtr _packet) { //what is _packet??

  datapacket_t datapacket;
  datapacket.first =  _packet->srcId;
  datapacket.second = _packet->srcSequenceNumber;    
  auto rangeIt = countAckMap.find(datapacket); 
  //todo if not end
  cout << "time " << Scheduler::now() << " node " << hostNode->getId() << " src " << rangeIt->first.first << " seqno " << rangeIt->first.second  << " rangeSmall " << rangeIt->second << "\n" << endl; 
}

//==============================================================================
//
//          backoffDynamicRingSendingEvent  (class)
//
//==============================================================================

backoffDynamicRingSendingEvent::backoffDynamicRingSendingEvent(simulationTime_t _t, Node* _node,int _packetId) : Event(_t){
  eventType = EventType::BACKOFF_SENDING_EVENT;
  node = _node;
  packetId = _packetId;
  //         packet = _packet;
}

const string backoffDynamicRingSendingEvent::getEventName(){
  return "backoff Ring Sending Event";
}

void backoffDynamicRingSendingEvent::consume() {
  //           cout<< " backoffDynamicRingSendingEvent::consume() BEGIN  on node "<< node->getId() << " for packet " << packetId << endl;
  node->getRoutingAgent()->processBackoffSending(packetId);
}
