/*
 * Copyright (C) 2017-2021 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */


#include <iostream>
#include <cassert>
#include "slr-ring-routing-agent.h"
#include "../../scheduler.h"
#include "../../utils.h"
#include "../../node.h"
#include "../../world.h"
#include "../../events.h"
#include <assert.h>

using namespace std;


using SLRInitialisationGenerationEvent = CallMethodEvent<SLRDynamicRingRoutingAgent,
							 &SLRDynamicRingRoutingAgent::processPacketGenerationEvent,
							 EventType::SLR_INITIALISATION_PACKET_GENERATION>;



//==============================================================================
//
//          SLRDynamicRingRoutingAgent  (class)
//
//==============================================================================

ofstream SLRDynamicRingRoutingAgent::SLRPositionsFile;
mt19937_64 *SLRDynamicRingRoutingAgent::forwardingRNG =  new mt19937_64(1);
int SLRDynamicRingRoutingAgent::forwardedSLRBeacons = 0;


SLRDynamicRingRoutingAgent::SLRDynamicRingRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  needToSendControl = true;
  update = false;
  rangeBig = ScenarioParameters::getRangeBig();
  //rangeSmall = ScenarioParameters::getRangeSmall();

  SLRCoordX = -1;
  SLRCoordY = -1;
  SLRCoordZ = -1;

  initialisationStarted = false;

  anchorID = -1;
  if (_hostNode->getId() == anchor1id)
    anchorID = 0;
  else if (_hostNode->getId() == anchor2id)
    anchorID = 1;
  else if (_hostNode->getId() == anchor3id)
    anchorID = 2;
  if ( anchorID == 0 ){
    simulationTime_t t = Scheduler::now()+200000000000;
    initialisationStarted = true;
    Scheduler::getScheduler().schedule(new SLRInitialisationGenerationEvent(t,this));
  }
}

SLRDynamicRingRoutingAgent::~SLRDynamicRingRoutingAgent() {
  if ( SLRPositionsFile.is_open() ) {
    SLRPositionsFile << hostNode->getId() << " " << SLRCoordX << " " << SLRCoordY << " " << SLRCoordZ << endl;
  }

  //     if (SLRCoordX == -1 || SLRCoordY == -1 || SLRCoordZ == -1 ){
  //      cout << " Destructor on node " << hostNode->getId() << " SLR coord X : " << SLRCoordX << " Y : " << SLRCoordY << " Z : " << SLRCoordZ << endl;
  //     }

  aliveAgents--;
  if ( aliveAgents == 0 ) {
    //      TODO : add global information
    cout << " SLR Routing with ring:: forwarded packets: " << forwardedDataPackets << endl;
    cout << " SLR Routing with ring:: forwarded beacons: " <<  forwardedSLRBeacons << endl;
  }
  //    if ( SLRCoordX == -1 || SLRCoordY == -1  || SLRCoordZ == -1 ){
  //// if ( hostNode->getId() < 30  ){
  //        cout << "SLRDynamicRingRoutingAgent destructor on Node " << hostNode->getId() ;
  //        cout << " SLR coord " << SLRCoordX << " " << SLRCoordY << " " << SLRCoordZ ;
  //        cout << endl;
  //    }
}

void SLRDynamicRingRoutingAgent::initializeAgent() {
  slrOpenFile (SLRPositionsFile);
  readCommRangeSetupAttribute ();
  readAnchorAttributes ();
  readRangeAttributes ();
}

void SLRDynamicRingRoutingAgent::receivePacketFromNetwork(PacketPtr _packet) {
  LogSystem::EventsLogOutput.log( LogSystem::routingRCV, Scheduler::now(),hostNode->getId(),(int) _packet->type,_packet->flowId,_packet->flowSequenceNumber, _packet->modifiedBitsPositions.size());

  //          cout << Scheduler::getScheduler().now() << " SLRDynamicRingRoutingAgent::receivePacket on Node " << hostNode->getId();
  //         cout << " srcId:" << _packet->srcId << " srcSeq:" << _packet->srcSequenceNumber;
  //         cout << " dstId:" << _packet->dstId << " port:" << _packet->port ;
  //         cout << " size:" << _packet->size;
  //         cout << " flowId:" << _packet->flowId << " flowSeq:" << _packet->flowSequenceNumber;
  //            cout << endl;

  //                   cout << " Forwarding packet type " << (int)_packet->type << " at " << Scheduler::now() <<  endl;
  //                   if ((int)_packet->type == 4) getchar();

  bool alreadySeen = false;
  bool forwardbeacon = false;
  auto it =alreadyForwardedPackets.find(_packet->flowId);
  if (it != alreadyForwardedPackets.end() ){ // Flow already followed
    if ( it->second >= _packet->flowSequenceNumber){ // The last seen packet from is flow is newer than the received one
      alreadySeen = true;
    }
  }

  //                 if (_packet->needAck && isAck(_packet)){

  if (  _packet->type == PacketType::D1_DENSITY_INIT && !alreadySeen ) {
    hostNode->dispatchPacketToApplication(_packet);
    //PacketPtr packetClone(new Packet(_packet));
    PacketPtr packetClone(_packet->clone());
    alreadyForwardedPackets.insert(pair<int,int>(_packet->flowId,_packet->flowSequenceNumber));
    hostNode->enqueueOutgoingPacket(packetClone);
    LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) packetClone->type,packetClone->flowId,packetClone->flowSequenceNumber);
    return;
  }
  if ( _packet->type == PacketType::D1_DENSITY_PROBE) {
    hostNode->dispatchPacketToApplication(_packet);
    return;
  }

  if (_packet->type == PacketType::SLR_BEACON ){
    update = SLRUpdate(_packet);
    if (update) forwardbeacon = true;
    if ( anchorID != -1   &&  _packet->anchorID != anchorID && !initialisationStarted ){
      initialisationStarted = true;
      simulationTime_t t = Scheduler::now()+100000000*anchorID;
      Scheduler::getScheduler().schedule(new SLRInitialisationGenerationEvent(t,this));
    }
  }

  if (_packet->type == PacketType::DATA && SLRDestinationReach(_packet)) {
    hostNode->dispatchPacketToApplication(_packet);
    //         return;
  }

  if (!alreadySeen){
    if(_packet->type == PacketType::DATA){
      if ( SLRForward(_packet) && isCloser(_packet) ){
	auto it =alreadyForwardedPackets.find(_packet->flowId);
	if (it != alreadyForwardedPackets.end()){
	  it->second = _packet->flowSequenceNumber ;
	}
	else{
	  alreadyForwardedPackets.insert(pair<int,int>(_packet->flowId,_packet->flowSequenceNumber));
	}
	//                      cout << " forwarding packet " <<  _packet->flowSequenceNumber << " on node " << hostNode->getId() << endl;
	if(amIOnRing(_packet))
	  forwardDataPackets(_packet);
      }
    }
  }


  if (_packet->type == PacketType::CONTROL_1) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// transmiter already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoBig = _packet->ctrl1SequenceNumber;
    } 
    else {
      // transmitter seen for the first time
      controSeqNo_t control;
      control.controlSeqNoBig = _packet->ctrl1SequenceNumber;
      control.controlSeqNoSmall = -1;
      amIonRingMap.insert(pair<int,controSeqNo_t>(_packet->transmitterId, control));
    }	
  }

  if (_packet->type == PacketType::CONTROL_2) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
    } 
    else {
      // seen for the first time
      controSeqNo_t control;
      control.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
      control.controlSeqNoBig = -1;
      amIonRingMap.insert(pair<int,controSeqNo_t>(_packet->transmitterId, control));	
    }	
  }



  if ( forwardbeacon ) {
    PacketPtr packetClone(_packet->clone()); //DO WE DO THIS FOR BEACON??
    packetClone->tx_SLRX = SLRCoordX;
    packetClone->tx_SLRY = SLRCoordY;
    packetClone->tx_SLRZ = SLRCoordZ;
    if ( update ) {packetClone->anchorDist++;}
    hostNode->enqueueOutgoingPacket(packetClone);
    LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) packetClone->type,packetClone->flowId,packetClone->flowSequenceNumber);
    forwardedSLRBeacons++;
  }

}

//amIOnRing
bool SLRDynamicRingRoutingAgent::amIOnRing( PacketPtr _packet ) {
  auto trIt = amIonRingMap.find(_packet->transmitterId);
  if (trIt == amIonRingMap.end()) 
    return false;

  else {
    if ((trIt->second.controlSeqNoBig > trIt->second.controlSeqNoSmall) || (trIt->second.controlSeqNoBig != -1 && trIt->second.controlSeqNoSmall == -1)){
      //cout << hostNode->getId() << " onringof " << _packet->transmitterId << endl;
      //cout << hostNode->getId() << " onring of " << amIonRingMap.size() << " transmitters" << endl; //journal
      return true;}
    else
      return false;
  }
}

//amIOnRingCount
int SLRDynamicRingRoutingAgent::amIOnRingCount() {
	int i =0;
	for(map<int,controSeqNo_t> ::const_iterator itt = amIonRingMap.begin(); itt != amIonRingMap.end(); ++itt)
		    {
		  if ((itt->second.controlSeqNoBig > itt->second.controlSeqNoSmall) || (itt->second.controlSeqNoBig != -1 && itt->second.controlSeqNoSmall == -1))
		  i++;
		    }
	return i;
}

void SLRDynamicRingRoutingAgent::forwardDataPackets( PacketPtr _packet ) {
  if (hostNode->getId() == 0) rangeSmall = 0;

  float optimal_NbRingNeighborsPerHop = 150.0;
  if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

    //rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));
	
    //journal if 3D??
    rangeSmall = pow(((1 - optimal_NbRingNeighborsPerHop/hostNode->getEstimatedNeighbours()) * pow(ScenarioParameters::getRangeBig(),3)),1.0/3);

  else rangeSmall = 0;
     
  cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

  if (needToSendControl) { //before the very first data transmission
    PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
    controlBig->setCommunicationRange(rangeBig);
    controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
    hostNode->enqueueOutgoingPacket(controlBig);
    PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
    controlSmall->setCommunicationRange(rangeSmall);
    controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
    hostNode->enqueueOutgoingPacket(controlSmall);	
    needToSendControl = false;
  }

  PacketPtr packetClone(_packet->clone());
  packetClone->tx_SLRX = SLRCoordX;
  packetClone->tx_SLRY = SLRCoordY;
  packetClone->tx_SLRZ = SLRCoordZ;
  if ( update ) {packetClone->anchorDist++;}
  cout << hostNode->getId() << " onring of " << amIOnRingCount() << " transmitters" << endl; //journal
  hostNode->enqueueOutgoingPacket(packetClone);
  LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) packetClone->type,packetClone->flowId,packetClone->flowSequenceNumber);
  forwardedDataPackets++;

  datapacket_t datapacket;
  datapacket.first =  packetClone->srcId;
  datapacket.second = packetClone->srcSequenceNumber;    
  auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
  if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
    auto timerEvent = new SLRDynamicRingAckEvent(Scheduler::now(), this, packetClone); 
    Scheduler::getScheduler().schedule(timerEvent);
  }                 
}

void SLRDynamicRingRoutingAgent::processAckTimerEvent(PacketPtr _packet) { //what is _packet??

if (hostNode->getId() == 0) return;

  datapacket_t datapacket;
  datapacket.first =  _packet->srcId;
  datapacket.second = _packet->srcSequenceNumber;    
  auto rangeIt = countAckMap.find(datapacket); 
  //todo if not end
  cout << "time " << Scheduler::now() << " node " << hostNode->getId() << " src " << rangeIt->first.first << " seqno " << rangeIt->first.second  << " rangeSmall " << rangeIt->second << "\n" << endl; 
}

// TODO: Delta functions ignore their Z parameter. SLR May not work in 3D !


bool SLRDynamicRingRoutingAgent::isCloser(PacketPtr _packet){
  bool res = false;

  int dstX,dstY,dstZ;
  int srcX,srcY,srcZ;
  int txX,txY,txZ;
  int X,Y,Z;

  /*
  // Florian's version areWeCloserThanPacketSender
  dstX = _packet->dst_SLRX; //added _ before packet
  dstY = _packet->dst_SLRY;
  dstZ = _packet->dst_SLRZ;
  txX = _packet->tx_SLRX;
  txY = _packet->tx_SLRY;
  txZ = _packet->tx_SLRZ;

  X = SLRCoordX; //slrx;
  Y = SLRCoordY; //slry;
  Z = SLRCoordZ; //slrz;

  // one. if the packet sender is closer in one single dimension,
  // we are not closer.
  if ((std::abs(dstX - txX) < std::abs(dstX - X)) ||
    (std::abs(dstY - txY) < std::abs(dstY - Y)) ||
    (std::abs(dstZ - txZ) < std::abs(dstZ - Z)))
    res = false;
  // two. if the packet sender is in the same zone, we are not closer.
  else if (txX == X && txY == Y && txZ == Z)
    res = false;
  // three. anybody left must be equal or closer on all dimensions, and
  // actually closer in at least one of them.
  else
    res = true;
 */
//journal
  dstX = _packet->dst_SLRX;
  dstY = _packet->dst_SLRY;
  if ( _packet->dst_SLRZ == -1 && _packet->dst_SLRX != -1 && _packet->dst_SLRY != -1 ){
    dstZ = _packet->dst_SLRX;
  } else {
    dstZ = _packet->dst_SLRZ;
  }

  srcX = _packet->src_SLRX;
  srcY = _packet->src_SLRY;
  if ( _packet->src_SLRZ == -1 && _packet->src_SLRX != -1 && _packet->src_SLRY != -1 ){
    srcZ = _packet->src_SLRX;
  } else {
    srcZ = _packet->src_SLRZ;
  }

  txX = _packet->tx_SLRX;
  txY = _packet->tx_SLRY;
  if ( _packet->tx_SLRZ == -1 && _packet->tx_SLRX != -1 && _packet->tx_SLRY != -1 ){
    txZ = _packet->tx_SLRX;
  } else {
    txZ = _packet->tx_SLRZ;
  }

  X = SLRCoordX;
  Y = SLRCoordY;

  if ( SLRCoordZ == -1 && SLRCoordX != -1 && SLRCoordY != -1 ){
    Z=X;
  } else {
    Z = SLRCoordZ;
  }

  if ( dstX == -1 && dstY == -1 && dstZ == -1){
    int dist,disttx;
    dist = abs(X - srcX) + abs(Y - srcY) ;
    disttx = abs(txX - srcX) + abs(txY - srcY) ;

    if (dist > disttx){
      res = true;
    }

//       if (!res){
//       printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
//       printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
//       printf("  X   %d  Y   %d  Z   %d \n", X,Y,Z);
//       printf("dist    %d \n",dist);
//       printf("disttx  %d \n",disttx);
//       printf("%d \n",res);
//       getchar();
//       }
//       res = true;
    return res;
  }

  if ( (X == dstX) && (Y==dstY) && (Z==dstZ) ){
    res = true;
    //     return true;
  }

  if ( (srcX <= dstX) && (X > txX) ) {
    res = true;
    //     return true;
  }
  if ( (srcX > dstX) && (X < txX) ) {
    res = true;
    //     return true;
  }

  if ( (srcY < dstY) && (Y > txY) ) {
    res = true;
    //     return true;
  }
  if ( (srcY > dstY) && (Y < txY) ) {
    res = true;
    //     return true;
  }

  if ( (srcZ < dstZ) && (Z > txZ) ) {
    res = true;
    //     return true;
  }
  if ( (srcZ > dstZ) && (Z < txZ) ) {
    res = true;
    //     return true;
  }//journal

  //  if (res) cout << " isCloser on node " << hostNode->getId() << " res = " << res << endl;

  //   printf(" dstX %d dstY %d dstZ %d \n", dstX,dstY,dstZ);
  //   printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
  //   printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
  //   printf("  X   %d  Y   %d  Z   %d \n", X,Y,Z);
  //
  //   getchar();
  return res;
}


bool SLRDynamicRingRoutingAgent::SLRForward(PacketPtr _packet, int m )
{

  int dstX,dstY,dstZ;
  int srcX,srcY,srcZ;
  int X,Y,Z;

  dstX = _packet->dst_SLRX;
  dstY = _packet->dst_SLRY;
  if ( _packet->dst_SLRZ == -1 && _packet->dst_SLRX != -1 && _packet->dst_SLRY != -1 ){
    dstZ = _packet->dst_SLRX;
  }
  else {
    dstZ = _packet->dst_SLRZ;
  }

  srcX = _packet->src_SLRX;
  srcY = _packet->src_SLRY;

  if ( _packet->src_SLRZ == -1 && _packet->src_SLRX != -1 && _packet->src_SLRY != -1 ){
    srcZ = _packet->src_SLRX;
  }
  else {
    srcZ = _packet->src_SLRZ;
  }

  X = SLRCoordX;
  Y = SLRCoordY;

  if ( SLRCoordZ == -1 && SLRCoordX != -1 && SLRCoordY != -1 ){
    Z=X;
  }
  else {
    Z = SLRCoordZ;
  }

  if (dstX == -1 && dstY == -1 && dstZ == -1) {
    return true;
  }

  int deltaA = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z);
  int deltaB = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z);

  int deltaAX = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z);
  int deltaAY = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y-m,Z);
  int deltaAZ = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y-m,Z);

  int deltaBX = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z);
  int deltaBY = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z-m);
  int deltaBZ = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z-m);

  //Node is on the line between src and dst
  bool onTheLine= ((deltaA*deltaAX <=0 || deltaA*deltaAY <= 0 || deltaA*deltaAZ <= 0) && (deltaB*deltaBX <=0 || deltaB*deltaBY <= 0 || deltaB*deltaBZ <= 0) );
  bool onSegment = ( ((X-dstX)*(X-srcX) <= 0)  && ((Y-dstY)*(Y-srcY) <= 0) && ((Z-dstZ)*(Z-srcZ) <= 0) );

  //    cout << " Routin on node " << hostNode->getId() << " SLR : " << onTheLine << " " << onSegment << endl;
  /*
    printf(" dstX %d dstY %d dstZ %d \n", dstX,dstY,dstZ);
    printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
    printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
    printf("    X %d    Y %d Z    %d \n", X,Y,Z);
    getchar();
  */
  return onTheLine && onSegment;
}


bool SLRDynamicRingRoutingAgent::SLRForward(PacketPtr _packet)
{
  //The default parameter is 1;
  return SLRForward(_packet,ScenarioParameters::getSlrPathWidth());
}


bool SLRDynamicRingRoutingAgent::SLRDestinationReach(PacketPtr _packet)
{

  if (_packet->dstId == -1 ){
    //       cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")"  << endl;
    LogSystem::EventsLogOutput.log(LogSystem::slrBroadcastReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber,SLRCoordX,SLRCoordY,SLRCoordZ);
  }
  //   if (hostNode->getId() == _packet->dstId) {
  //         cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")"  << endl;
  // //         getchar();
  //       return true;
  //   }

  if (_packet->dst_SLRX == SLRCoordX && _packet->dst_SLRY == SLRCoordY && _packet->dst_SLRZ == SLRCoordZ ) {
    cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")"  << endl;
    LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber,_packet->type);

    //     getchar();
    return true;
  }
  return false;
}

void SLRDynamicRingRoutingAgent::receivePacketFromApplication(PacketPtr _packet) {
  if (_packet->type == PacketType::D1_DENSITY_PROBE || _packet->type == PacketType::D1_DENSITY_INIT){ //on all nodes approx
    hostNode->enqueueOutgoingPacket(_packet); 
  }

  else if (_packet->type == PacketType::SLR_BEACON){  //DO WE DO THIS FOR BEACON??
    PacketPtr newPacket(_packet->clone());
    newPacket->tx_SLRX = SLRCoordX;
    newPacket->tx_SLRY = SLRCoordY;
    newPacket->tx_SLRZ = SLRCoordZ;
    if ( update ) {newPacket->anchorDist++;} 
    hostNode->enqueueOutgoingPacket(newPacket);
    forwardedSLRBeacons++;
  }

  else if (_packet->type == PacketType::DATA){

  float optimal_NbRingNeighborsPerHop = 150.0;
  if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

    //rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));
	
    //journal if 3D??
    rangeSmall = pow(((1 - optimal_NbRingNeighborsPerHop/hostNode->getEstimatedNeighbours()) * pow(ScenarioParameters::getRangeBig(),3)),1.0/3);

  else rangeSmall = 0;
     
  cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

    if (needToSendControl) { //before the very first data transmission
      PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
      controlBig->setCommunicationRange(rangeBig);
      controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlBig);
      PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
      controlSmall->setCommunicationRange(rangeSmall);
      controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlSmall);	
      needToSendControl = false;
    }
    _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
    _packet->src_SLRX = SLRCoordX;
    _packet->src_SLRY = SLRCoordY;
    _packet->src_SLRZ = SLRCoordZ;

    _packet->tx_SLRX = SLRCoordX;
    _packet->tx_SLRY = SLRCoordY;
    _packet->tx_SLRZ = SLRCoordZ;

    //TODO : how to find values here?
    if (_packet->type == PacketType::DATA && _packet->dstId != -1 ){
      Node* dst = World::getNode(_packet->dstId);
      SLRDynamicRingRoutingAgent* dstAgent = dynamic_cast<SLRDynamicRingRoutingAgent*> (dst->getRoutingAgent());

      if ( dstAgent == nullptr ) {
	cerr << " Error in setting SLR destination to a packet. The targeted destination may not run SLR " << endl;

      }
      else {
	_packet->dst_SLRX = dstAgent->getSLRX();
	_packet->dst_SLRY = dstAgent->getSLRY();
	_packet->dst_SLRZ = dstAgent->getSLRZ();
      }
    }
    else if (_packet->type == PacketType::DATA  && _packet->dstId == -1 ){
      _packet->dst_SLRX = -1;
      _packet->dst_SLRY = -1;
      _packet->dst_SLRZ = -1;
    }
		cout << hostNode->getId() << " onring of " << amIOnRingCount() << " transmitters" << endl; //journal
    hostNode->enqueueOutgoingPacket(_packet);
    datapacket_t datapacket;
    datapacket.first =  _packet->srcId;
    datapacket.second = _packet->srcSequenceNumber;    
    auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
    if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
      auto timerEvent = new SLRDynamicRingAckEvent(Scheduler::now(), this, _packet); 
      Scheduler::getScheduler().schedule(timerEvent);
    }        
  }
}
/////////////////////////////////////////

bool SLRDynamicRingRoutingAgent::SLRUpdate(PacketPtr _packet)
{
  int anchorId = _packet->anchorID;
  int anchorDist =  _packet->anchorDist;
  bool update=false;

  SLRDynamicRingRoutingAgent* routing = static_cast<SLRDynamicRingRoutingAgent*> (hostNode->getRoutingAgent());
  //   cout << " SLR update on node " << hostNode->getId() << " by anchor " << anchorId << endl;

  switch ( anchorId )
    {
    case 0:
      if (anchorDist < routing->getSLRX() || routing->getSLRX() == -1){
	routing->setSLRX(anchorDist);
	//                 routing->setSLRZ(anchorDist);
	update=true;
      }
      break;
    case 1:
      if (anchorDist < routing->getSLRY() || routing->getSLRY() == -1){
	routing->setSLRY(anchorDist);
	update=true;
      }
      break;
    case 2:
      if (anchorDist < routing->getSLRZ() || routing->getSLRZ() == -1){
	routing->setSLRZ(anchorDist);
	update=true;
      }
      break;
    default:
      cerr << " *** ERROR *** SLR_BEACON received from unknown anchor " << anchorId << endl;
      exit (EXIT_FAILURE);
    }

  return update;
}

void SLRDynamicRingRoutingAgent::processPacketGenerationEvent() {
  cout << " anchor " << anchorID << " on node " << hostNode->getId() << " sent a beacon" << endl;
  // SLR beacons have a -1 flow id
  PacketPtr p(new Packet(PacketType::SLR_BEACON, 100, hostNode->getId(), -1, -1,-1, 0,anchorID,1));
  SLRUpdate(p);
  hostNode->getRoutingAgent()->receivePacketFromApplication(p);
}

