/*
 * Copyright (C) 2017-2022 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */


#include <iostream>
#include <cassert>
#include "slr-routing-agent.h"
#include "../scheduler.h"
#include "../utils.h"
#include "../node.h"
#include "../world.h"
#include "../events.h"

using namespace std;


using SLRSetupPacketEvent = CallMethodEvent<SLRRoutingAgent,
  &SLRRoutingAgent::processSetupPacketEvent,
  EventType::SLR_SETUP_PACKET_GENERATION>;

using SLRSendEvent = CallMethodArgEvent<SLRRoutingAgent,PacketPtr,
  &SLRRoutingAgent::processPacketRetransmission,EventType::SLR_PACKET_RETRANSMISSION_DEST_ZONE>;

using SLRBackoffSendEvent = CallMethodArgEvent<SLRRoutingAgent,int,
  &SLRRoutingAgent::processPacketBackoffRetransmission,EventType::SLR_BACKOFF_PACKET_RETRANSMISSION>;

//==============================================================================
//
//          SLRRoutingAgent  (class)
//
//==============================================================================

ofstream SLRRoutingAgent::SLRPositionsFile;
mt19937_64 *SLRRoutingAgent::backoffRNG;
int SLRRoutingAgent::forwardedSLRBeacons = 0;
map<int,SLRPosition> SLRRoutingAgent::SLRPositionsMapFromFile;
int SLRRoutingAgent::redundancy;
float SLRRoutingAgent::dataBackoffMultiplier;
bool SLRRoutingAgent::useCounterBasedForwarding;
bool SLRRoutingAgent::useDeviation;
simulationTime_t SLRRoutingAgent::firstBeaconStartTime;
simulationTime_t SLRRoutingAgent::intervalBetweenBeaconStart;


SLRRoutingAgent::SLRRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  SLRCoordX = -1;
  SLRCoordY = -1;
  SLRCoordZ = -1;

  setupStarted = false;

  anchorID = -1;
  if (_hostNode->getId() == anchor1id)
    anchorID = 0;
  else if (_hostNode->getId() == anchor2id)
    anchorID = 1;
  else if (_hostNode->getId() == anchor3id)
    anchorID = 2;

  auto it = SLRPositionsMapFromFile.find( _hostNode->getId() );
  if ( it != SLRPositionsMapFromFile.end() ) {
    SLRCoordX = it->second.x;
    SLRCoordY = it->second.y;
    SLRCoordZ = it->second.z;
  } else {
    if ( anchorID == 0 ){
      setupStarted = true;
      Scheduler::getScheduler().schedule (new SLRSetupPacketEvent (Scheduler::now()+firstBeaconStartTime, this));
    }
  }
}

SLRRoutingAgent::~SLRRoutingAgent() {
  if ( SLRPositionsFile.is_open() ) {
    SLRPositionsFile << hostNode->getId() << " " << SLRCoordX << " " << SLRCoordY << " " << SLRCoordZ << endl;
  }

  aliveAgents--;
  if ( aliveAgents == 0 ) {
    //      TODO : add global information
    cout << " SLR Routing:: forwarded packets: " << forwardedDataPackets << endl;
    cout << " SLR Routing:: forwarded beacons: " <<  forwardedSLRBeacons << endl;
  }

  if ( waitingPacket.size() > 0) {
    cerr << "Waitingpacket should be empty on node " << hostNode->getId() << endl;
  }
}

void SLRRoutingAgent::initializeAgent() {
  backoffRNG =  new mt19937_64( ScenarioParameters::getBackoffRNGSeed() );

  if ( ScenarioParameters::getLoadSLRPositionsFromFile() ) {
    string filename = ScenarioParameters::getScenarioDirectory() + "/" + ScenarioParameters::getPrefix() + "SLRPositions" + ScenarioParameters::getDefaultExtension();
    std::ifstream infile(filename);
    if ( infile.fail() ) {
      cerr << "*** WARNING Loading SLR Positions from " << filename << " failed with the following error: " << strerror(errno) << endl;
      cerr << "***         SLR setup phase will be used instead." << endl;
    } else {
      std::string line;
      int counter = 0;
      while (std::getline(infile, line)) {
        std::istringstream iss(line);
        struct SLRPosition pos;
        int n;
        if (!(iss >> n >> pos.x >> pos.y >> pos.z)) { break; } else {
          SLRPositionsMapFromFile.insert(pair<int, SLRPosition>(n,pos));
          counter++;
        }
      }
      cout << "*** INFO " << counter << " SLR positions read from " << filename << endl;
    }
    infile.close();
  }

  slrOpenFile (SLRPositionsFile);
  readCommRangeSetupAttribute ();
  readAnchorAttributes ();
  readBackoffAttributes();
  readTimingAttributes();

  redundancy = ScenarioParameters::getBackoffRedundancy();
  dataBackoffMultiplier = ScenarioParameters::getBackoffMultiplier();
}


void SLRRoutingAgent::receivePacketFromNetwork(PacketPtr _packet) {
  LogSystem::EventsLogOutput.log( LogSystem::routingRCV, Scheduler::now(),hostNode->getId(),(int) _packet->type,_packet->flowId,_packet->flowSequenceNumber, _packet->modifiedBitsPositions.size());

  bool alreadySeen = false;
  bool update = false;
  bool forward = false;
  auto it =alreadyForwardedPackets.find(_packet->flowId);
  if (it != alreadyForwardedPackets.end() ){ // Flow already followed
    if ( it->second >= _packet->flowSequenceNumber){ // The last seen packet from is flow is newer than the received one
      alreadySeen = true;
    }
  }

  //                 if (_packet->needAck && isAck(_packet)){

  if ( _packet->type == PacketType::D1_DENSITY_INIT && !alreadySeen ) {
    hostNode->dispatchPacketToApplication(_packet);
    //PacketPtr newPacket(new Packet(_packet));
    PacketPtr newPacket(_packet->clone());
    alreadyForwardedPackets.insert(pair<int,int>(_packet->flowId,_packet->flowSequenceNumber));
    hostNode->enqueueOutgoingPacket(newPacket);
    LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) newPacket->type,newPacket->flowId,newPacket->flowSequenceNumber);
    return;
  }
  if (_packet->type == PacketType::D1_DENSITY_PROBE) {
    hostNode->dispatchPacketToApplication(_packet);
    return;
  }
  if (_packet->type == PacketType::SLR_BEACON ){
    update = SLRUpdate(_packet);
    if (update) forward = true;
    if ( anchorID != -1 &&  _packet->anchorID != anchorID && !setupStarted ){
      setupStarted = true;
      simulationTime_t t = Scheduler::now()+intervalBetweenBeaconStart*anchorID;
      Scheduler::getScheduler().schedule(new SLRSetupPacketEvent(t,this));
    }
  }

  simulationTime_t retransmitDelayAtDestinationZone = 0;

  if (_packet->type == PacketType::DATA && SLRDestinationReach(_packet)) {
    hostNode->dispatchPacketToApplication(_packet);
    //         return;

    // Used with sleep mechanism
    // When reaching the destination zone, the destination node may be sleeping
    // We want to retransmit the packet at the end of this node awaken period, in hope the destination node will be awake and catch it.
    //
    SleepingNode* sleepingHostNode = dynamic_cast<SleepingNode*>(hostNode);
    if ( sleepingHostNode ) {  // if sleeping is used
      simulationTime_t transmissionDate = Scheduler::now() - (Scheduler::now() % SleepingNode::ts) + sleepingHostNode->awakenStart + sleepingHostNode->awakenDuration;
      if (Scheduler::now()%SleepingNode::ts < sleepingHostNode->awakenStart) {
        // the awaken duration spread in 1 TS or in 2 TS
        transmissionDate -= SleepingNode::ts;
        assert (transmissionDate >= Scheduler::now());
      }
      retransmitDelayAtDestinationZone = transmissionDate - Scheduler::now();
      assert (retransmitDelayAtDestinationZone >= 0);
    }
  }

  // If I already saw this packet REDUNDANCY or more times, I prevent any retransmission already scheduled
  //
  // Also further copies of this packet will not be retransmitted either, as
  // alreadySeen list should prevent it.

  SleepingNode* sleepingHostNode = dynamic_cast<SleepingNode*>(hostNode);
  if ( useCounterBasedForwarding ) {
    for (auto it = waitingPacket.begin(); it!=waitingPacket.end();it++){
      if (_packet->type == PacketType::DATA && (it->second.p)->flowId == _packet->flowId && (it->second.p)->flowSequenceNumber == _packet->flowSequenceNumber) {
        if ( (useDeviation && isAck(_packet)) || (!useDeviation && implicitAck(_packet)) ) {
          if (it->second.counter >= redundancy) {
            // At the destination zone, all nodes should repeat the packet, to help the destination node to receive it.
            if ( !( sleepingHostNode && SLRDestinationReach(_packet)) ) {
              waitingPacket.erase(it);
            }
          } else {
            it->second.counter++;
          }
          break;
        }
      }
    }
  }

  if ( (!alreadySeen && _packet->type == PacketType::DATA)){
    if ( SLRForward(_packet) && isCloser(_packet) ){
      forward = true;
      auto it =alreadyForwardedPackets.find(_packet->flowId);
      if (it != alreadyForwardedPackets.end()){
        it->second = _packet->flowSequenceNumber ;
      }
      else{
        alreadyForwardedPackets.insert(pair<int,int>(_packet->flowId,_packet->flowSequenceNumber));
      }
      // cout << " forwarding packet " <<  _packet->flowSequenceNumber << " on node " << hostNode->getId() << endl;
    }
  }

  if ( forward ) {
    uniform_real_distribution<float> distrib(0, 1);
    // float forwardingProbabilityNumber = distrib(*forwardingRNG);
    //
    // int wantedForwarder = 40;
    // if ( _packet->type == PacketType::DATA) wantedForwarder = 15;

    // float forwadingProbability;
    // if ( hostNode->getEstimatedNeibhbours() == -1 ){
    //   forwadingProbability = 1;
    // } else {
    // forwadingProbability = ((float)wantedForwarder)/(float)hostNode->getEstimatedNeibhbours();
    // }

    bool retransmit = true;
    // bool retransmit = forwardingProbabilityNumber < forwadingProbability ;
    if (retransmit) {
      PacketPtr newPacket(_packet->clone());
      newPacket->tx_SLRX = SLRCoordX;
      newPacket->tx_SLRY = SLRCoordY;
      newPacket->tx_SLRZ = SLRCoordZ;
      if ( update ) {newPacket->anchorDist++;}

      if ( newPacket->type == PacketType::DATA) {
        newPacket->hopCount++;
      }

      // deviation
      if ( useDeviation && _packet->type == PacketType::DATA) {
        int mcr = ScenarioParameters::getMaxConcurrentReceptions();
        int receptCount = hostNode->getConcurrentReceptionCount();
        float congestionLevel = (float)receptCount / mcr;
        int deviateThresh = 50;
        int convergeThresh = 50;
        if (100 * congestionLevel > deviateThresh) {
          newPacket->deviation++;
          // cout << "deviating with " << congestionLevel << " to thresh " << deviateThresh << ". new level: " << packet->deviation << "\n";
        } else if (100 * congestionLevel < convergeThresh && newPacket->deviation > 1) {
          newPacket->deviation--;
          newPacket->src_SLRX = SLRCoordX;
          newPacket->src_SLRY = SLRCoordY;
          newPacket->src_SLRZ = SLRCoordZ;
        }
      }

      //if ( retransmitDelayAtDestinationZone == 0 ) {
      if (newPacket->type == PacketType::SLR_BEACON || !useCounterBasedForwarding){
        hostNode->enqueueOutgoingPacket(newPacket);
      } else {
        if ( _packet->type == PacketType::DATA) {
          struct backoffRoutingCounter info;
          info.p = newPacket;
          info.counter = 1;
          waitingPacket.insert(pair<int,backoffRoutingCounter>(newPacket->packetId,info));
        }

        simulationTime_t backoff = getBackoffDynamic(hostNode, newPacket, newPacket->type == PacketType::DATA ? dataBackoffMultiplier : 1, backoffRNG);
        SleepingNode* sleepingHostNode = dynamic_cast<SleepingNode*>(hostNode);
        if (sleepingHostNode) {
          backoff -= backoff % SleepingNode::ts; // Ts multiple for SleepingNode
        }

        //
        // Work in progress code to had controlled spread when using sleep
        //
        //uniform_int_distribution<int> intDdistrib(-1000, 1000);
        //int backoff2 = intDdistrib(*backoffRNG);
        //backoff2 *= 1;
        //backoff += backoff2;
        //if (backoff < 0) backoff = 0;

        auto sendEvent = new SLRBackoffSendEvent(Scheduler::now() + backoff + retransmitDelayAtDestinationZone, this, newPacket->packetId);
        Scheduler::getScheduler().schedule (sendEvent);
      }

      LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now()+retransmitDelayAtDestinationZone,hostNode->getId(),(int) newPacket->type,newPacket->flowId,newPacket->flowSequenceNumber);

      if (newPacket->type == PacketType::DATA) {
        forwardedDataPackets++;
      }
      else if (newPacket->type == PacketType::SLR_BEACON){
        forwardedSLRBeacons++;
      }
    }
  }
}

// TODO: Delta functions ignore their Z parameter. SLR May not work in 3D !


bool SLRRoutingAgent::isCloser(PacketPtr _packet){
  bool res = false;

  int dstX,dstY,dstZ;
  int srcX,srcY,srcZ;
  int txX,txY,txZ;
  int X,Y,Z;

/*  
  // Florian's version areWeCloserThanPacketSender
  dstX = _packet->dst_SLRX; //added _ before packet
  dstY = _packet->dst_SLRY;
  dstZ = _packet->dst_SLRZ;
  txX = _packet->tx_SLRX;
  txY = _packet->tx_SLRY;
  txZ = _packet->tx_SLRZ;

  X = SLRCoordX; //slrx;
  Y = SLRCoordY; //slry;
  Z = SLRCoordZ; //slrz;

  // one. if the packet sender is closer in one single dimension,
  // we are not closer.
  if ((std::abs(dstX - txX) < std::abs(dstX - X)) ||
    (std::abs(dstY - txY) < std::abs(dstY - Y)) ||
    (std::abs(dstZ - txZ) < std::abs(dstZ - Z)))
    res = false;
  // two. if the packet sender is in the same zone, we are not closer.
  else if (txX == X && txY == Y && txZ == Z)
    res = false;
  // three. anybody left must be equal or closer on all dimensions, and
  // actually closer in at least one of them.
  else
    res = true;
*/ 
 //journal
  dstX = _packet->dst_SLRX;
  dstY = _packet->dst_SLRY;
  if ( _packet->dst_SLRZ == -1 && _packet->dst_SLRX != -1 && _packet->dst_SLRY != -1 ){
    dstZ = _packet->dst_SLRX;
  } else {
    dstZ = _packet->dst_SLRZ;
  }

  srcX = _packet->src_SLRX;
  srcY = _packet->src_SLRY;
  if ( _packet->src_SLRZ == -1 && _packet->src_SLRX != -1 && _packet->src_SLRY != -1 ){
    srcZ = _packet->src_SLRX;
  } else {
    srcZ = _packet->src_SLRZ;
  }

  txX = _packet->tx_SLRX;
  txY = _packet->tx_SLRY;
  if ( _packet->tx_SLRZ == -1 && _packet->tx_SLRX != -1 && _packet->tx_SLRY != -1 ){
    txZ = _packet->tx_SLRX;
  } else {
    txZ = _packet->tx_SLRZ;
  }

  X = SLRCoordX;
  Y = SLRCoordY;

  if ( SLRCoordZ == -1 && SLRCoordX != -1 && SLRCoordY != -1 ){
    Z=X;
  } else {
    Z = SLRCoordZ;
  }

  if ( dstX == -1 && dstY == -1 && dstZ == -1){
    int dist,disttx;
    dist = abs(X - srcX) + abs(Y - srcY) ;
    disttx = abs(txX - srcX) + abs(txY - srcY) ;

    if (dist > disttx){
      res = true;
    }

//       if (!res){
//       printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
//       printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
//       printf("  X   %d  Y   %d  Z   %d \n", X,Y,Z);
//       printf("dist    %d \n",dist);
//       printf("disttx  %d \n",disttx);
//       printf("%d \n",res);
//       getchar();
//       }
//       res = true;
    return res;
  }

  if ( (X == dstX) && (Y==dstY) && (Z==dstZ) ){
    res = true;
    //     return true;
  }

  if ( (srcX <= dstX) && (X > txX) ) {
    res = true;
    //     return true;
  }
  if ( (srcX > dstX) && (X < txX) ) {
    res = true;
    //     return true;
  }

  if ( (srcY < dstY) && (Y > txY) ) {
    res = true;
    //     return true;
  }
  if ( (srcY > dstY) && (Y < txY) ) {
    res = true;
    //     return true;
  }

  if ( (srcZ < dstZ) && (Z > txZ) ) {
    res = true;
    //     return true;
  }
  if ( (srcZ > dstZ) && (Z < txZ) ) {
    res = true;
    //     return true;
  } //journal

  //  if (res) cout << " isCloser on node " << hostNode->getId() << " res = " << res << endl;

  //   printf(" dstX %d dstY %d dstZ %d \n", dstX,dstY,dstZ);
  //   printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
  //   printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
  //   printf("  X   %d  Y   %d  Z   %d \n", X,Y,Z);
  //
  //   getchar();
  return res;
  
}


bool SLRRoutingAgent::implicitAck(PacketPtr _packet) {
  //
  // Should return true if the packet has been sent by a node on the path while being at the same coordinates than the current node or closer to the destination
  //

  int dstX = _packet->dst_SLRX;
  int dstY = _packet->dst_SLRY;
  int dstZ = _packet->dst_SLRZ;

  int srcX = _packet->src_SLRX;
  int srcY = _packet->src_SLRY;
  int srcZ = _packet->src_SLRZ;

  int txX = _packet->tx_SLRX;
  int txY = _packet->tx_SLRY;
  int txZ = _packet->tx_SLRZ;

  int X = SLRCoordX;
  int Y = SLRCoordY;
  int Z = SLRCoordZ;

  if ( SLRForward( _packet) ) {
    //
    // current node is on the path of this packet
    //

    if ( X == txX && Y == txY && Z == txZ ) {
      return true;
    }

    if ( dstX != -1 || dstY != -1 || dstZ != -1 ) {
      //
      // The packet is not broadcasted (dst != -1,-1,-1)
      // I have to verify it came from downstream

      if ( (srcX < dstX) && (txX > X) ) {
        return true;
      }
      if ( (srcX > dstX) && (txX < X) ) {
        return true;
      }

      if ( (srcY < dstY) && (txY > Y) ) {
        return true;
      }
      if ( (srcY > dstY) && (txY < Y) ) {
        return true;
      }

      if ( (srcZ < dstZ) && (txZ > Z) ) {
        return true;
      }
      if ( (srcZ > dstZ) && (txZ < Z) ) {
        return true;
      }
    }
  }
  return false;
}

// for deviation
bool SLRRoutingAgent::isAck(PacketPtr _packet) // The packet is coming from "further"
{
  int X = SLRCoordX;
  int Y = SLRCoordY;
  int Z = SLRCoordZ;

  int txX = _packet->tx_SLRX;
  int txY = _packet->tx_SLRY;
  int txZ = _packet->tx_SLRZ;

  int srcX = _packet->src_SLRX;
  int srcY = _packet->src_SLRY;
  int srcZ = _packet->src_SLRZ;

  int dstX = _packet->dst_SLRX;
  int dstY = _packet->dst_SLRY;
  int dstZ = _packet->dst_SLRZ;

//     if (SLRForward(_packet, _packet->deviation)){

  if ( (X == dstX) && (Y == dstY) && (Z == dstZ) ){
    return true;
  }

  if ( (srcX < dstX) && (X < txX) ) {
    return true;
  }
  if ( (srcX > dstX) && (X > txX) ) {
    return true;
  }

  if ( (srcY < dstY) && (Y < txY) ) {
    return true;
  }
  if ( (srcY > dstY) && (Y > txY) ) {
    return true;
  }

  if ( (srcZ < dstZ) && (Z < txZ) ) {
    return true;
  }
  if ( (srcZ > dstZ) && (Z > txZ) ) {
    return true;
  }


//     }

  return false;
}

bool SLRRoutingAgent::SLRForward(PacketPtr _packet, int m )
{

  int dstX,dstY,dstZ;
  int srcX,srcY,srcZ;
  int X,Y,Z;

  dstX = _packet->dst_SLRX;
  dstY = _packet->dst_SLRY;
  if ( _packet->dst_SLRZ == -1 && _packet->dst_SLRX != -1 && _packet->dst_SLRY != -1 ){
    dstZ = _packet->dst_SLRX;
  } else {
    dstZ = _packet->dst_SLRZ;
  }

  srcX = _packet->src_SLRX;
  srcY = _packet->src_SLRY;

  if ( _packet->src_SLRZ == -1 && _packet->src_SLRX != -1 && _packet->src_SLRY != -1 ){
    srcZ = _packet->src_SLRX;
  } else {
    srcZ = _packet->src_SLRZ;
  }

  X = SLRCoordX;
  Y = SLRCoordY;

  if ( SLRCoordZ == -1 && SLRCoordX != -1 && SLRCoordY != -1 ){
    Z=X;
  } else {
    Z = SLRCoordZ;
  }

  if (dstX == -1 && dstY == -1 && dstZ == -1) {
    return true;
  }

  int deltaA = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z);
  int deltaB = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z);

  int deltaAX = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z);
  int deltaAY = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y-m,Z);
  int deltaAZ = DeltaA(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y-m,Z);

  int deltaBX = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z);
  int deltaBY = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X,Y,Z-m);
  int deltaBZ = DeltaB(dstX,dstY,dstZ,srcX,srcY,srcZ, X-m,Y,Z-m);

  //Node is on the line between src and dst
  bool onTheLine = (deltaA*deltaAX <= 0 || deltaA*deltaAY <= 0 || deltaA*deltaAZ <= 0) && (deltaB*deltaBX <= 0 || deltaB*deltaBY <= 0 || deltaB*deltaBZ <= 0);
  bool onSegment = (X-dstX)*(X-srcX) <= 0 && (Y-dstY)*(Y-srcY) <= 0 && (Z-dstZ)*(Z-srcZ) <= 0;
  //journal nothing wrong with these equations (just like slr article)
  

  //    cout << " Routin on node " << hostNode->getId() << " SLR : " << onTheLine << " " << onSegment << endl;
  /*
    printf(" dstX %d dstY %d dstZ %d \n", dstX,dstY,dstZ);
    printf(" srcX %d srcY %d srcZ %d \n", srcX,srcY,srcZ);
    printf("  txX %d  txY %d  txZ %d \n", txX,txY,txZ);
    printf("    X %d    Y %d Z    %d \n", X,Y,Z);
    getchar();
  */
  return onTheLine && onSegment;
}


bool SLRRoutingAgent::SLRForward(PacketPtr _packet)
{
  return SLRForward(_packet,_packet->deviation);
}


bool SLRRoutingAgent::SLRDestinationReach(PacketPtr _packet)
{
  if (_packet->dstId == -1 ){
//       cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")"  << endl;
    LogSystem::EventsLogOutput.log(LogSystem::slrBroadcastReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber,SLRCoordX,SLRCoordY,SLRCoordZ);
  }

  //   TODO : DO we want a zone reception or a node reception ?

  //   if (hostNode->getId() == _packet->dstId) {
  //         cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")"  << endl;
  //       return true;
  //   }

  if (_packet->dst_SLRX == SLRCoordX && _packet->dst_SLRY == SLRCoordY && _packet->dst_SLRZ == SLRCoordZ ) {
    cout << " DESTINATION REACHED on node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << _packet->flowSequenceNumber << "(" << _packet->flowId << ")" << " " << "at time" << " " << Scheduler::now()<< endl;
    LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber,_packet->type,SLRCoordX,SLRCoordY,SLRCoordZ);
    return true;
  }
  return false;
}

void SLRRoutingAgent::receivePacketFromApplication(PacketPtr _packet) {
  if ( useCounterBasedForwarding ) _packet->disableBackoff = true;

  _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());

  _packet->src_SLRX = SLRCoordX;
  _packet->src_SLRY = SLRCoordY;
  _packet->src_SLRZ = SLRCoordZ;

  _packet->tx_SLRX = SLRCoordX;
  _packet->tx_SLRY = SLRCoordY;
  _packet->tx_SLRZ = SLRCoordZ;

  //TODO : how to find values here?
  if (_packet->type == PacketType::DATA && _packet->dstId != -1 ){
    Node* dst = World::getNode(_packet->dstId);
    SLRRoutingAgent* dstAgent = dynamic_cast<SLRRoutingAgent*> (dst->getRoutingAgent());

    if ( dstAgent == nullptr ) {
      cerr << " Error in setting SLR destination to a packet. The targeted destination may not run SLR " << endl;

    } else {
      _packet->dst_SLRX = dstAgent->getSLRX();
      _packet->dst_SLRY = dstAgent->getSLRY();
      _packet->dst_SLRZ = dstAgent->getSLRZ();
    }
  } else if (_packet->type == PacketType::DATA  && _packet->dstId == -1 ){
    _packet->dst_SLRX = -1;
    _packet->dst_SLRY = -1;
    _packet->dst_SLRZ = -1;
  }

  hostNode->enqueueOutgoingPacket(_packet);
}

bool SLRRoutingAgent::SLRUpdate(PacketPtr _packet)
{
  int anchorId = _packet->anchorID;
  int anchorDist =  _packet->anchorDist;
  bool update=false;

  SLRRoutingAgent* routing = static_cast<SLRRoutingAgent*> (hostNode->getRoutingAgent());
  //   cout << " SLR update on node " << hostNode->getId() << " by anchor " << anchorId << endl;

  switch ( anchorId )
  {
  case 0:
    if (anchorDist < routing->getSLRX() || routing->getSLRX() == -1){
      routing->setSLRX(anchorDist);
//                 routing->setSLRZ(anchorDist);
      update=true;
    }
    break;
  case 1:
    if (anchorDist < routing->getSLRY() || routing->getSLRY() == -1){
      routing->setSLRY(anchorDist);
      update=true;
    }
    break;
  case 2:
    if (anchorDist < routing->getSLRZ() || routing->getSLRZ() == -1){
      routing->setSLRZ(anchorDist);
      update=true;
    }
    break;
  default:
    cerr << " *** ERROR *** SLR_BEACON received from unknown anchor " << anchorId << endl;
    exit (EXIT_FAILURE);
  }

  return update;
}

void SLRRoutingAgent::processSetupPacketEvent() {
  cout << "SLR: anchor " << anchorID << " sent a beacon" << endl;
  // SLR beacons have a -1 flow id
  PacketPtr p(new Packet(PacketType::SLR_BEACON, 100, hostNode->getId(), -1, -1,-1, 0,anchorID,1));
  SLRUpdate(p);
  hostNode->getRoutingAgent()->receivePacketFromApplication(p);
}

void SLRRoutingAgent::processPacketRetransmission (PacketPtr packet){
  //cout << "The packet is retransmitted with the node : " << " " << hostNode->getId() << " " << "\n";

  //cout << " Node " << hostNode->getId() << "(" << SLRCoordX << "," << SLRCoordY << "," << SLRCoordZ << ")" << " packet (flow) " << "(" << packet->flowId << ")"  <<" " << "current time " << " " << Scheduler::now() <<" "<< "Waiting Time " << " " << waitingTime << endl;
  hostNode->enqueueOutgoingPacket (packet);
}

void SLRRoutingAgent::processPacketBackoffRetransmission(int packetId){
     auto it = waitingPacket.find(packetId);
  if ( it != waitingPacket.end() && it->second.p->type == PacketType::DATA ) {
    //
    // if the packet is still in the waiting list
    // i.e. I have not see more than REDUNDANCY copies
    // then I send it now
    //
    //cout << " ++ forward seq " << it->second.p->flowSequenceNumber << " by node " << hostNode->getId()<< " at " << Scheduler::getScheduler().now() << endl;
    hostNode->enqueueOutgoingPacket(it->second.p);
    LogSystem::EventsLogOutput.log( LogSystem::routingSND, Scheduler::now(),hostNode->getId(),(int) it->second.p->type,it->second.p->flowId,it->second.p->flowSequenceNumber);
//         LogSystem::EventsLogOutput.log( LogSystem::memoryTrace, Scheduler::now(),hostNode->getId(),"-",it->second.p->flowId,it->second.p->flowSequenceNumber);

    forwardedDataPackets++;
    waitingPacket.erase(it);
    //       string filename = ScenarioParameters::getScenarioDirectory()+"/dataSent.data";
    //       ofstream dataSent(filename,std::ofstream::app);
    //       dataSent<< ScenarioParameters::getDefaultBeta() << " " <<hostNode->getId() << " "<< Scheduler::now() << endl;
    //       dataSent.close();
    //
    //
    //       string filename2 = ScenarioParameters::getScenarioDirectory()+"/memoryTrace.data";
    //       ofstream memoryTrace(filename2,std::ofstream::app);
    //       memoryTrace << ScenarioParameters::getDefaultBeta() <<" " << hostNode->getId() <<  " - "<< Scheduler::now() << endl;
    //       memoryTrace.close();
  }
}

void SLRRoutingAgent::readBackoffAttributes() {
  tinyxml2::XMLElement *slrElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());

  std::string enabled;
  ScenarioParameters::queryStringAttr(slrElement, "useCounterBasedForwarding", enabled, false, nullptr, "false", "counter-based forwarding is not used");
  useCounterBasedForwarding = (enabled == "true" ? true : false);
  ScenarioParameters::queryStringAttr(slrElement, "useDeviation", enabled, false, nullptr, "false", "deviation-based forwarding is not used");
  useDeviation = (enabled == "true" ? true : false);
}

void SLRRoutingAgent::readTimingAttributes() {
  tinyxml2::XMLElement *slrElement = ScenarioParameters::getXMLRootNode()->FirstChildElement("routing")->FirstChildElement(ScenarioParameters::getRoutingAgentName().c_str());

  ScenarioParameters::queryLongAttr(slrElement, "firstBeaconStartTime_fs", "firstBeaconStartTime_ns", firstBeaconStartTime, false, nullptr, 200000000000, "");
  ScenarioParameters::queryLongAttr(slrElement, "intervalBetweenBeaconStart_fs", "intervalBetweenBeaconStart_ns", intervalBetweenBeaconStart, false, nullptr, 100000000, "");
}
